# Generated by roxygen2: do not edit by hand

export("%::%")
export(as_interface)
export(assert_implements)
export(assert_trait)
export(has_trait)
export(impl_trait)
export(implements)
export(interface_report)
export(interface_requirement)
export(interface_requirements)
export(missing_requirements)
export(new_interface)
export(new_trait)
export(trait_assoc_const)
export(trait_assoc_type)
export(trait_call)
export(trait_method)
export(trait_methods)
export(trait_report)
